'''Wyvern is a Py3k wrapper for the Twitter REST API'''

__author__ = 'Daniel Moran <danxmoran@gmail.com>'
__version__ = '0.1.0'

from .endpoints import rest_endpoints as endpoint_table
from .endpoints import rest_errors as error_table
from .endpoints import base_url

try:
	import requests
	from requests_oauthlib import OAuth1
except ImportError:
	err_message = \
		'Wyvern requires \"requests\" and \"requests_oauthlib\" to work. ' + \
		'Use \"pip install requests requests_oauthlib\"'
	raise ImportError(err_message)
import re, requests

class WyvernError(Exception):
	'''General exception class'''
	def __init__(self, status_code, general_explanation,
	             detailed_explanation, internal_err_code):
		self.status_code = status_code
		self.general_explanation = general_explanation
		self.detailed_explanation = detailed_explanation
		self.internal_err_code = internal_err_code
		err_message = str(self)
		super(WyvernError, self).__init__(err_message)

	def __str__(self):
		err_message = 'The following error occurred: ' +\
			'{} {} - {}'.format(self.status_code, self.general_explanation,
			                    self.detailed_explanation) + '.'
		return err_message

class OAuthError(WyvernError):
	'''Raised when a request is made with bad OAuth credentials'''
	def __str__(self):
		err_message = super(OAuthError, self).__str__() + \
			' Make sure your OAuth data is correct and try again.'
		return err_message

class WyvernUserError(WyvernError):
	'''Raised when the user makes an invalid call to the Twitter API'''
	def __str__(self):
		err_message = super(WyvernUserError, self).__str__() + \
			' Make sure your request parameters are correct and try again.'
		return err_message

class WyvernTwitterError(WyvernError):
	'''Raised when something breaks on Twitter's end'''
	def __str__(self):
		err_message = super(WyvernTwitterError, self).__str__() + \
			' This is Twitter\'s fault, try again in a few minutes.'
		return err_message

class WyvernUserRateLimitError(WyvernUserError):
	'''Raised when the user makes too many calls to the API in a given window'''
	def __init__(self, *params, reset_time):
		self.reset_time = reset_time
		super(WyvernUserRateLimitError, self).__init__(*params)

	def __str__(self):
		from datetime import timedelta
		import time

		wait_time = timedelta(seconds=int(self.reset_time) - time.time())
		super_message = super(WyvernUserRateLimitError, self).__str__()
		minutes = wait_time.seconds // 60
		seconds = wait_time.seconds % 60
		err_message = super_message[:len(super_message) - 1] + \
			' after {} minutes and {} seconds'.format(minutes, seconds)
		return err_message

class Wyvern(object):
	'''Client for the Twitter REST API'''
	########################### DOCUMENTATION HELPER ###########################
	@staticmethod
	def method_for_endpoint(endpoint):
		'''
		Prints to screen the name of the wyvern method corresponding to 
		the given Twitter API endpoint.

		:param endpoint: the string value of the endpoint to be queried

		Raises a ValueError if endpoint is not a valid Twitter API endpoint.
		'''
		def swap_format(endpoint):
			pattern = re.compile(r'^(.*?):(.+?)(/.*)*$')
			maybe_match = re.match(pattern, endpoint)
			if maybe_match:
				begin, var, end = maybe_match.groups()
				if not var and not end:
					return begin
				elif not end:
					end = ''
					return begin + '{' + var + '}' + end
			else:
				return endpoint

		if endpoint == 'account/settings':
			print('\n', 'That endpoint can be used with either GET or POST','\n')
			print('GET: my_account_settings')
			print('POST: update_my_account_settings\n')
		for func in endpoint_table:
			if endpoint_table[func]['endpoint'] == swap_format(endpoint):
				print('\n', func, '\n')
				return
		err_message = \
			'Endpoint {} does not exist. '.format(endpoint) + \
			'See <https://dev.twitter.com/docs/api/1.1> for ' + \
			'the list of endpoints currently supported by Twitter'
		raise ValueError(err_message)
	###########################################################################

	def __init__(self, consumer_key, consumer_secret, access_token,
	             access_token_secret, headers=None, proxies=None, version=1.1):
		self.version = version
		self.headers = headers or {'User-Agent': 'Wyvern/' + __version__}
		self.proxies = proxies
		self.oauth_header = OAuth1(client_key = consumer_key,
		                           client_secret = consumer_secret,
		                           resource_owner_key = access_token, 
		                           resource_owner_secret = access_token_secret,
		                           signature_type = 'auth_header')

		# Make sure OAuth data is correct before going any further
		verification_response = self.verify_my_credentials()

	def __getattr__(self, attr_name):
		'''
		General creator for calls to the API.

		Use Wyvern.method_for_endpoint(endpont) to find the actual method name
		for the endpoint you are trying to work with.  All parameters to API 
		calls must be given as keyword-arg pairs. Calls to endpoints which 
		contain a variable such as :id or :slug must also contain a keyword-arg 
		pairing for that variable.
		'''
		if attr_name.startswith('_'):
			raise AttributeError(attr_name)
		else:
			def api_caller(**call_params):
				try:
					api_endpoint = endpoint_table[attr_name]['endpoint']
					call_method = endpoint_table[attr_name]['method']
				except:
					err_message = \
						'Method {} does not exist. Use '.format(attr_name) + \
						'Wyvern.method_for_endpoint(endpoint) to find ' + \
						'the method you are looking for.'
					raise NameError(err_message)
				response = self._call_api(call_method, api_endpoint,
				                          **call_params)
				return self._decode_response(response)
			return api_caller

	def _call_api(self, call_method, api_endpoint, **call_params):
		'''
		Makes a call to the Twitter API.

		:param call_method: the HTTP method to use in the API call,
							either GET or POST
		:param api_endpoint: the endpoint of the Twitter API to call
		:param **call_params: the parameters to be passed in the call;
							  parameters differ from endpoint to endpoint

		Returns a :class Response object from the requests module
		containing data from the result of the call.
		'''
		var_pattern = re.compile(r'^.*?\{(.+?)\}.*$')
		if re.match(var_pattern, api_endpoint):
			var = re.match(var_pattern, api_endpoint).groups()[0]
			api_endpoint = api_endpoint.format(**{var: call_params.pop(var)})

		url = base_url.format(**{'version': self.version}) + \
			  api_endpoint + '.json'

		if call_method == 'GET':
			return requests.get(url,
			                    params=call_params,
			                    auth=self.oauth_header, 
			                    headers=self.headers, 
			                    proxies=self.proxies)
		else:
			return requests.post(url,
			                     data=call_params,
			                     auth=self.oauth_header, 
			                     headers=self.headers, 
			                     proxies=self.proxies)

	def _decode_response(self, response):
		'''
		Decodes a response received from a call to the Twitter API.

		:param response: the result of the API call

		Returns a :class dict containing the json-decoded content of the 
		response, or raises error message if the call did not complete 
		successfully.
		'''
		if response.status_code == 200:
			return response.json()
		else:
			error_code = response.status_code
			error_message = error_table[error_code]
			internal_message = response.json()['errors'][0]['message']
			internal_code = response.json()['errors'][0]['code']

			if error_code // 100 == 4:
				if error_code == 429:
					reset_time = response.headers['x-rate-limit-reset']
					raise WyvernUserRateLimitError(error_code, error_message,
					                               internal_message,
					                               internal_code,
					                               reset_time=reset_time)
				elif error_code == 401:
					raise OAuthError(error_code, error_message,
					                 internal_message, internal_code)
				else:
					raise WyvernUserError(error_code, error_message,
					                      internal_message, internal_code)
			else:
				raise WyvernTwitterError(error_code, error_message,
				                         internal_message, internal_code)
